from __future__ import print_function

import logging
import os

import ipdb
import psutil
import ray
import tensorflow_probability as tfp

from algorithms.position import position
from test_env.tactile_env import tactile_env
from utils import *


class tactile(position):
    def is_grasp_master(self):
        return False

    def custom_init_op(self, host=None, port=None):
        self.max_train_success_perc = 0
        self.build_grasp_placeholders_op()

        self.grasp_curr_obv_tr, self.grasp_bl_obv_tr = self.grasp_build_perception_net(g.grasp_cnn, self.grasp_img)

        self.grasp_curr_pi = self.grasp_build_pi_network_op(scope='%s_curr_grasp_pi' % g.grasp_cnn,
                                                            input=self.grasp_curr_obv_tr, act_ph=self.grasp_act_ph,
                                                            trainable=False)

        self.build_tactile_placeholders_op()
        self.tactile_curr_obv_tr = self.tactile_build_perception_net(scope='curr_tactile_perception',
                                                                     trainable=self.is_tactile_master())
        self.tactile_curr_pi = self.tactile_build_pi_network_op(scope='curr_tactile_pi', input=self.tactile_curr_obv_tr,
                                                                act_ph=self.tactile_act_on_ph if g.on_policy else
                                                                self.tactile_act_off_ph,
                                                                trainable=self.is_tactile_master())
        if self.is_tactile_master() and not self.eval:
            self.tactile_old_obv_tr = self.tactile_build_perception_net(scope='old_tactile_perception', trainable=False)
            self.tactile_old_pi = self.tactile_build_pi_network_op(scope='old_tactile_pi',
                                                                   input=self.tactile_old_obv_tr,
                                                                   act_ph=self.tactile_act_on_ph if g.on_policy else
                                                                   self.tactile_act_off_ph,
                                                                   trainable=False)

            self.tactile_q1_on_tr = self.build_v_network_op('q1', act_ph=self.tactile_act_on_ph, n_obv=False,
                                                            trainable=True) if g.off_policy else None
            self.tactile_q2_on_tr = self.build_v_network_op('q2', act_ph=self.tactile_act_on_ph, n_obv=False,
                                                            trainable=True) if g.off_policy else None
            self.tactile_q1_off_tr = self.build_v_network_op('q1', act_ph=self.tactile_act_off_ph, n_obv=False,
                                                             trainable=True, reuse=True) if g.off_policy else None
            self.tactile_q2_off_tr = self.build_v_network_op('q2', act_ph=self.tactile_act_off_ph, n_obv=False,
                                                             trainable=True, reuse=True) if g.off_policy else None
            self.tactile_curr_v_tr = self.build_v_network_op('vf', n_obv=False, trainable=True)
            if g.off_policy:
                self.tactile_next_v_target_tr = self.build_v_network_op('v_target', n_obv=True, trainable=False)
            else:
                self.tactile_next_v_target_tr = None
            self.tensorboard_keys += ['loss/v_loss_tr', 'loss/q1_loss_tr', 'loss/q2_loss_tr', 'loss/positive_mse_tr',
                                      'loss/negative_mse_tr'] if g.off_policy else ['loss/improved_tactile_variance',
                                                                                    'loss/v_loss_tr']
            with tf.variable_scope('update_ops'):
                self.update_old_pi_op = self.build_update_old_pi_op(curr_scope='curr_tactile', old_scope='old_tactile')
                if g.off_policy:
                    self.update_v_target_op = self.build_update_v_target_op(v_scope='vf', v_target_scope='v_target')
            with tf.variable_scope('train_ops'):
                self.train_pi_op, self.pi_loss_tr = self.build_pi_train_op(scope='curr_tactile',
                                                                           curr_pi=self.tactile_curr_pi,
                                                                           old_pi=self.tactile_old_pi,
                                                                           lr_ph=self.tactile_pi_lr_ph,
                                                                           v=self.tactile_curr_v_tr,
                                                                           q=self.tactile_q1_off_tr,
                                                                           lr_mult_ph=self.tactile_lr_mult_ph,
                                                                           adv_ph=self.tactile_adv_ph)
                self.train_v_op, self.v_loss_tr = self.build_v_train_op(self.tactile_curr_v_tr, self.tactile_v_lr_ph,
                                                                        q1=self.tactile_q1_off_tr,
                                                                        q2=self.tactile_q2_off_tr,
                                                                        curr_pi=self.tactile_curr_pi, scope='vf',
                                                                        vtarg_ph=self.tactile_v_target_ph)
                if g.off_policy:
                    self.train_q1_op, self.q1_loss_tr = self.build_q_train_op(q=self.tactile_q1_on_tr,
                                                                              next_obv_v_target=self.tactile_next_v_target_tr,
                                                                              r_ph=self.tactile_r_ph, scope='q1',
                                                                              q_lr_ph=self.tactile_q_lr_ph,
                                                                              done_ph=self.tactile_done_ph,
                                                                              qtarg_ph=self.tactile_v_target_ph)
                    self.train_q2_op, self.q2_loss_tr = self.build_q_train_op(q=self.tactile_q2_on_tr,
                                                                              next_obv_v_target=self.tactile_next_v_target_tr,
                                                                              r_ph=self.tactile_r_ph, scope='q2',
                                                                              q_lr_ph=self.tactile_q_lr_ph,
                                                                              done_ph=self.tactile_done_ph,
                                                                              qtarg_ph=self.tactile_v_target_ph)

            self.build_text_ops()
            self.build_tactile_text_ops()
            self.build_tactile_histogram_ops()
            self.tensorboard_keys += ['experience/num_train_traj', 'experience/num_ts',
                                      'experience/effective_timesteps_so_far', 'experience/real_timesteps_so_far',
                                      'seconds', 'loss/pi_loss_tr', 'rewards/max_train_success_perc',
                                      'rewards/train_success_perc', 'rewards/train_regrasp_success_perc',
                                      'quality/max_test_multi_novel_success_perc',
                                      'quality/max_test_multi_seen_success_perc',
                                      'quality/max_test_single_novel_success_perc',
                                      'quality/max_test_single_seen_success_perc',
                                      'quality/test_multi_seen_success_perc', 'quality/test_multi_novel_success_perc',
                                      'quality/test_single_seen_success_perc', 'quality/test_single_novel_success_perc',
                                      'rewards/regrasp_penalty', 'rewards/best_avg_train_r', 'rewards/avg_train_r',
                                      'lr/tactile_v_lr', 'lr/tactile_pi_lr', 'quality/baseline_r',
                                      'rewards/max_outperformance', 'rewards/outperformance', 'env/num_objects',
                                      'lr/lr_mult', 'system/memory', 'experience/dynamic_tactile_delta',
                                      'experience/pi_replay_buffer_size', 'experience/v_replay_buffer_size',
                                      'experience/v_replay_buffer_pos_size', 'experience/v_replay_buffer_neg_size',
                                      'experience/valid_buffer_size']
            for obj in g.seen_objs:
                self.tensorboard_keys.append('per_obj_accuracy/%s' % obj)
            self.pi_replay_buffer = Queue(maxsize=g.replay_buffer_queue_size)
            self.v_replay_buffer_pos = Queue(maxsize=g.mini_bs * g.replay_buffer_num_batch_mult)
            self.v_replay_buffer_neg = Queue(maxsize=g.mini_bs * g.replay_buffer_num_batch_mult)
            self.v_replay_buffer = Queue(maxsize=g.replay_buffer_num_batch_mult)
        else:
            self.env = tactile_env(host, port, eval=self.eval, real=self.real, id=self.id)

    def build_tactile_placeholders_op(self):
        with tf.variable_scope('inputs'):
            self.is_tactile_training_ph = tf.placeholder(tf.bool, name='is_tactile_training_ph')

            self.contacts_obv_ph = tf.placeholder(tf.float32, shape=[None, 96 * g.history], name='contacts_obv_ph')
            self.positions_obv_ph = tf.placeholder(tf.float32, shape=[None, 8 * g.history], name='positions_obv_ph')
            self.cartesians_obv_ph = tf.placeholder(tf.float32, shape=[None, 96 * 3 * g.history],
                                                    name='cartesians_obv_ph')
            self.velocities_obv_ph = tf.placeholder(tf.float32, shape=[None, 8 * g.history], name='velocities_obv_ph')
            self.forces_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.history], name='forces_obv_ph')
            self.torques_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.history], name='torques_obv_ph')
            self.efforts_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.history], name='efforts_obv_ph')

            if g.tactile_delta:
                self.delta_contacts_obv_ph = tf.placeholder(tf.float32, shape=[None, 96 * g.delta_history],
                                                            name='delta_contacts_obv_ph')
                self.delta_positions_obv_ph = tf.placeholder(tf.float32, shape=[None, 8 * g.delta_history],
                                                             name='delta_positions_obv_ph')
                self.delta_cartesians_obv_ph = tf.placeholder(tf.float32, shape=[None, 96 * 3 * g.delta_history],
                                                              name='delta_cartesians_obv_ph')
                self.delta_velocities_obv_ph = tf.placeholder(tf.float32, shape=[None, 8 * g.delta_history],
                                                              name='delta_velocities_obv_ph')
                self.delta_forces_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.delta_history],
                                                          name='delta_forces_obv_ph')
                self.delta_torques_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.delta_history],
                                                           name='delta_torques_obv_ph')
                self.delta_efforts_obv_ph = tf.placeholder(tf.float32, shape=[None, 3 * g.delta_history],
                                                           name='delta_efforts_obv_ph')

            for name in g.tactile_obv_names:
                if g.off_policy:
                    tensor_name = '%s_next_obv_ph' % name
                    setattr(self, tensor_name, tf.placeholder(tf.float32, shape=[None, int(
                        getattr(self, '%s_obv_ph' % name).get_shape()[1])], name=tensor_name))

            self.tactile_done_ph = tf.placeholder(tf.float32, shape=[None], name='tactile_done_ph')
            self.force_open_ph = tf.placeholder(tf.float32, shape=[None], name='force_open_ph')
            self.force_lift_ph = tf.placeholder(tf.float32, shape=[None], name='force_lift_ph')
            self.tactile_r_ph = tf.placeholder(tf.float32, shape=[None], name='tactile_r_ph')
            self.tactile_nonlogit_dim = 4 + g.regrasp + g.tactile_regrasp
            self.tactile_logit_dim = g.tactile_regrasp * g.num_adjust_logits
            self.tactile_act_dim = 4 + g.regrasp + g.tactile_regrasp * 3
            self.tactile_act_on_ph = tf.placeholder(tf.float32, shape=[None, self.tactile_act_dim],
                                                    name='tactile_act_on_ph')
            self.tactile_act_off_ph = tf.placeholder(tf.float32, shape=[None, self.tactile_act_dim],
                                                     name='tactile_act_off_ph')
            self.success_rate_per_obj_values_ph = tf.placeholder(tf.float32, shape=[None],
                                                                 name='success_rate_per_obj_values_ph')
            if self.is_tactile_master():
                self.tactile_pi_lr_ph = tf.placeholder(tf.float32, name='tactile_pi_lr_ph')
                self.tactile_v_lr_ph = tf.placeholder(tf.float32, name='tactile_v_lr_ph')
                self.tactile_q_lr_ph = tf.placeholder(tf.float32, name='tactile_q_lr_ph')
                self.tactile_lr_mult_ph = tf.placeholder(tf.float32, name='tactile_lr_mult_ph')
                self.tactile_adv_ph = tf.placeholder(tf.float32, [None], name='tactile_adv_ph')
                self.tactile_v_target_ph = tf.placeholder(tf.float32, (None, 1), name='tactile_v_target_ph')

    def build_v_network_op(self, scope, n_obv, trainable, reuse=False, act_ph=None):
        with tf.variable_scope(scope):
            features = []
            initializer = normc_initializer(1)

            for name in g.tactile_obv_names:
                if getattr(g, 'tactile_%s' % name):
                    if g.direct_v_network_concat:
                        features.append(getattr(self, ('%s_next_obv_ph' if n_obv else '%s_obv_ph') % name))
                    else:
                        features.append(dnn(getattr(self, ('%s_next_obv_ph' if n_obv else '%s_obv_ph') % name),
                                            output_size=g.tactile_embedding_size, scope=name, n_layers=g.qv_n_layers,
                                            size=g.qv_hid_fc_size, trainable=self.is_tactile_master() and trainable,
                                            hid_init=initializer, final_init=initializer,
                                            final_activation_fn_str='relu', reuse=reuse, hid_act_fn_str='None'))
            if act_ph is not None:
                if g.direct_v_network_concat:
                    features.append(self.tactile_act_on_ph)
                else:
                    features.append(dnn(self.tactile_act_on_ph, output_size=g.tactile_embedding_size, scope='action',
                                        n_layers=g.qv_n_layers, size=g.qv_hid_fc_size,
                                        trainable=self.is_tactile_master() and trainable, hid_init=initializer,
                                        final_init=initializer, final_activation_fn_str='relu', reuse=reuse,
                                        hid_act_fn_str='None'))

            feature = tf.concat(features, axis=1)
            return dnn(feature, output_size=1, scope='dnn', n_layers=g.qv_n_layers, size=g.qv_hid_fc_size,
                       trainable=self.is_tactile_master() and trainable, hid_init=initializer, final_init=initializer,
                       final_activation_fn_str='relu', reuse=reuse)

    def tactile_build_perception_net(self, scope, trainable):
        with tf.variable_scope(scope):
            features = []
            initializer = normc_initializer(1)

            for name in g.tactile_obv_names:
                if getattr(g, 'tactile_%s' % name):
                    features.append(
                        dnn(getattr(self, '%s_obv_ph' % name), output_size=g.tactile_embedding_size, scope=name,
                            n_layers=g.tactile_pi_n_layers, size=g.tactile_pi_hid_fc_size, trainable=trainable,
                            hid_init=initializer, final_init=initializer, final_activation_fn_str='tanh'))
                    features.append(dnn(getattr(self, 'delta_%s_obv_ph' % name), output_size=g.tactile_embedding_size,
                                        scope='delta_' + name, n_layers=g.tactile_pi_n_layers,
                                        size=g.tactile_pi_hid_fc_size, trainable=trainable, hid_init=initializer,
                                        final_init=initializer, final_activation_fn_str='tanh'))

            feature = tf.concat(features, axis=1)

            non_logits = dnn(feature, output_size=self.tactile_nonlogit_dim, scope='nonlogits',
                             n_layers=g.tactile_pi_n_layers, size=g.tactile_pi_hid_fc_size, trainable=trainable,
                             hid_init=initializer, final_init=initializer, final_activation_fn_str='sigmoid')
            pos_logits = dnn(feature, output_size=self.tactile_logit_dim, scope='pos_logits',
                             n_layers=g.tactile_pi_n_layers, size=g.tactile_pi_hid_fc_size, trainable=trainable,
                             hid_init=initializer, final_init=initializer, final_activation_fn_str='tanh')
            ori_logits = dnn(feature, output_size=self.tactile_logit_dim, scope='ori_logits',
                             n_layers=g.tactile_pi_n_layers, size=g.tactile_pi_hid_fc_size, trainable=trainable,
                             hid_init=initializer, final_init=initializer, final_activation_fn_str='tanh')
            return tf.concat((non_logits, pos_logits, ori_logits), axis=1)

    def tactile_build_pi_network_op(self, scope, input, act_ph, trainable):
        pi = {}
        adjust_dist_dist, pos_adjust_direction_dist, ori_adjust_direction_dist, adjust_dist_mean = None, None, None, \
                                                                                                   None
        with tf.variable_scope(scope + '/pi'):
            if g.regrasp:
                bernoulli_probs = input[:, :5]
                if g.tactile_regrasp:
                    adjust_dist_mean = input[:, 5:6]
                    pi['adjust_dist_log_std_raw'] = tf.get_variable('adjust_dist_log_std',
                                                                    shape=[int(adjust_dist_mean.shape[-1])],
                                                                    trainable=trainable,
                                                                    initializer=normc_initializer(0.1))
                    pi['adjust_dist_log_std'] = pi['adjust_dist_log_std_raw'] - g.log_std_offset
                    pi['adjust_dist_std'] = tf.exp(pi['adjust_dist_log_std'])
                    adjust_dist_dist = tfp.distributions.MultivariateNormalDiag(loc=adjust_dist_mean,
                                                                                scale_diag=pi['adjust_dist_std'])
                    pos_adjust_direction_logits = input[:, 6:6 + g.num_adjust_logits]
                    pos_adjust_direction_dist = tfp.distributions.MultivariateNormalDiag(
                        # adjust_dist_dist, pos_adjust_direction_dist won't be used later
                        loc=pos_adjust_direction_logits, scale_diag=pi['adjust_dist_std'])
                    ori_adjust_direction_logits = input[:, 6 + g.num_adjust_logits:6 + g.num_adjust_logits * 2]
                    ori_adjust_direction_dist = tfp.distributions.MultivariateNormalDiag(
                        loc=ori_adjust_direction_logits,
                        scale_diag=pi['adjust_dist_std'])
                        # ori_adjust_direction_dist will be used
            else:
                bernoulli_probs = input
            tactile_dist = tfp.distributions.Bernoulli(probs=bernoulli_probs)

            if g.tactile_regrasp:
                pi['sampled_action'] = tf.concat((tf.cast(tactile_dist.sample(), tf.float32), adjust_dist_dist.sample(),
                                                  tf.cast(ori_adjust_direction_dist.sample(), tf.float32),
                                                  tf.cast(ori_adjust_direction_dist.sample(), tf.float32)),
                                                 axis=1)
                pi['best_action'] = pi['sampled_action']
            else:
                pi['sampled_action'] = tactile_dist.sample()
                pi['best_action'] = bernoulli_probs > 0.5

            if self.is_tactile_master():
                pi['log_prob'] = self.compute_log_prob(act_ph=act_ph,
                                                       dists=[tactile_dist, adjust_dist_dist, pos_adjust_direction_dist,
                                                              ori_adjust_direction_dist])
        return pi

    def compute_log_prob(self, act_ph, dists):
        tactiLe_dist, adjust_dist_dist, pos_adjust_direction_dist, ori_adjust_direction_dist = dists
        terminate = act_ph[:, 3]

        tactile_dist_log_prob = tactiLe_dist.log_prob(act_ph[:, :5])
        finger_log_prob = tactile_dist_log_prob[:, :3]
        terminate_log_prob = tactile_dist_log_prob[:, 3]
        grasp_log_prob = (tf.reduce_sum(finger_log_prob, axis=1) * (1 - terminate) + terminate_log_prob) * (
                1 - self.force_open_ph)
        if g.regrasp:
            regrasp = (1 - (1 - act_ph[:, 4]) * (1 - self.force_open_ph))
            regrasp_log_prob = tactile_dist_log_prob[:, 4] * (1 - self.force_open_ph)
            log_prob = regrasp_log_prob + (1 - regrasp) * grasp_log_prob
            if g.tactile_regrasp:
                log_prob += regrasp * (adjust_dist_dist.log_prob(act_ph[:, 5:6]) + pos_adjust_direction_dist.log_prob(
                    act_ph[:, 6]) + ori_adjust_direction_dist.log_prob(act_ph[:, 7]))
        else:
            log_prob = grasp_log_prob
        return log_prob * (1 - self.force_lift_ph)

    def generate_grasp(self, test, multi, novel, num_objects, reinit_env=True, get_baseline=True, state_id=None):
        index = 0
        while True:
            if reinit_env:
                self.env.init_world(test=test, multi=multi, novel=novel, num_objects=num_objects)
                s = self.env.reset()
                while self.need_reinit_pybullet(s):
                    self.env.init_world(test=test, multi=multi, novel=novel, num_objects=num_objects)
                    s = self.env.reset()
            else:
                self.env.commander.restoreState(state_id)
                s = self.env.get_state(1, get_rgb=self.real)
            t = 1
            done = False
            curr_t_steps = 0
            try:
                while not done:
                    s['image_path'] = os.path.join(d.dirname, 'grasp_ts_%s_t_%s_id_%s.png' % (curr_t_steps, t, self.id))
                    cv2.imwrite(s['image_path'], s['image'])
                    action, obv = self.get_grasp_action(s, best=True)
                    a = self.get_grasp_action_dict(action[0])
                    if index is not None:
                        a['index'] = index
                    s, r, done, grasp_info = self.env.grasp_step(a, info={'s': s, 't': t},
                                                                 get_baseline=get_baseline and (
                                                                         self.t % g.check_tactile_baseline == 0))
                    curr_t_steps += 1
                return s
            except:
                traceback.print_exc()
                index += 1

    def train(self, num_objects):
        experience, traj, traj_r = [], [], []
        cum_r, num_traj, num_succ, curr_t_steps = [0] * 4

        self.env.obj_index = -1
        t = 1
        s = self.generate_grasp(test=False, multi=True, novel=False, num_objects=num_objects)
        state_id = s['state_id']
        regrasp = False
        while curr_t_steps < g.bs // g.num_actors or num_traj < (300 // g.num_actors):
            if curr_t_steps % 50 == 0:
                print('Progress: %s / %s' % (curr_t_steps, g.bs // g.num_actors))

            action, raw_act = self.get_tactile_action([s])
            next_s, r, done, info = self.env.tactile_step(self.get_tactile_action_dict(action[0]),
                                                          info={'t': t, 'eval': False, 's': s})
            if not done and 'visual_regrasp' in next_s:
                next_s = self.generate_grasp(test=False, multi=True, novel=False, num_objects=num_objects,
                                             get_baseline=False, state_id=state_id)
            if 'tactile_regrasp' in info:
                regrasp = True

            info['raw_act'] = raw_act[0]

            traj.append([s, action[0], r, next_s, done, info])
            cum_r += r

            if done:
                traj[-1][-1]['regrasp_success'] = (info['success'] and regrasp)
                traj = [self.save_memory(exp) for exp in traj]
                experience += traj
                traj = []
                s = self.generate_grasp(test=False, multi=True, novel=False, num_objects=num_objects)
                num_traj += 1
                traj_r.append(cum_r)
                cum_r = 0
                t = 1
                regrasp = False
            else:
                t += 1
                s = next_s

            curr_t_steps += 1
        return experience, num_traj, sum(traj_r), None

    def save_memory(self, traj):
        traj[3] = {}
        return traj

    def initialize_iterators(self, bs):
        pass

    def test_objects(self, novel, multi):
        successes = []
        baseline_r = []
        print('testing %s %s scenes...' % ('novel' if novel else 'seen', 'multi' if multi else 'single'))
        for episode in range(g.num_eval_episodes_per_actor):
            if episode % 1 == 0 and g.statistics:
                print('Progress: %s / %s' % (episode, self.id))
            s = self.generate_grasp(test=True, novel=novel, multi=multi, num_objects=None)
            state_id = s['state_id']
            baseline_r.append(s['baseline'])
            done = False
            t = 1
            while not done:
                action, raw_act = self.get_tactile_action([s], best=True)
                s, r, done, info = self.env.tactile_step(self.get_tactile_action_dict(action[0]),
                                                         info={'t': t, 'eval': True, 's': s})
                if not done and 'regrasp' in s:
                    s = self.generate_grasp(test=True, novel=novel, multi=multi, num_objects=None, get_baseline=False,
                                            state_id=state_id)
                t += 1
            successes.append(info['success'])
            if info['success'] < 1:
                print('failed!', self.env.commander.randomObjectName())
                cv2.imwrite(os.path.join(d.server_graspit_img_dir, '%s-%s-%s-%s-%s.png' % (
                    self.env.commander.randomObjectName(), self.id, episode, 'novel' if novel else 'seen',
                    'multi' if multi else 'single')), self.env.get_state(t, get_rgb=True)['rgb'])
            else:
                print('success!')
        return float(np.mean(successes)), float(np.mean(baseline_r)), float(np.mean(successes)) - float(
            np.mean(baseline_r))

    def get_tactile_action(self, obvs, best=False):
        if g.gpu:
            with tf.device('/gpu:0'):
                return self.get_tactile_action_raw(obvs, best)
        else:
            return self.get_tactile_action_raw(obvs, best)

    def get_tactile_action_raw(self, obvs, best):
        fd = {}
        for name in g.tactile_obv_names:
            if getattr(g, 'tactile_%s' % name):
                fd[getattr(self, '%s_obv_ph' % name)] = [obv[name] for obv in obvs]
                fd[getattr(self, 'delta_%s_obv_ph' % name)] = [obv['delta_' + name] for obv in obvs]
        act, obv = self.sess.run(
            [self.tactile_curr_pi['sampled_action' if not best else 'best_action'], self.tactile_curr_obv_tr],
            feed_dict=fd)
        return act, obv

    def get_tactile_action_dict(self, a):
        d = {}
        if g.regrasp:
            d['regrasp'] = a[4]
            if g.tactile_regrasp:
                d['pos_adjust_distance'] = a[5]
                d['pos_adjust_direction'] = a[6]
                d['ori_adjust_direction'] = a[7]
        d['finger1'] = a[0]
        d['finger2'] = a[1]
        d['finger3'] = a[2]
        d['terminate'] = a[3]
        d['delta'] = dynamicTactileDelta(
            accuracy=self.max_train_success_perc) if not self.eval else g.eval_tactile_delta
        return d

    def learn(self):
        if g.off_policy:
            self.sess.run(self.update_v_target_op)
            print([i.tolist() for i in
                   self.sess.run([self.tactile_q1_on_tr, self.tactile_v_target_ph], feed_dict=self.fd)])
            self.do_train([self.q1_loss_tr, self.train_q1_op, self.q2_loss_tr, self.train_q2_op], 'q',
                          buffer=self.v_replay_buffer, epochs=1, bs=None)
        self.do_train([self.v_loss_tr, self.train_v_op], 'v', buffer=self.v_replay_buffer, epochs=1, bs=None)
        self.do_train([self.pi_loss_tr, self.train_pi_op], 'pi', buffer=self.pi_replay_buffer, epochs=g.optim_epochs,
                      bs=None)
        self.sess.run([self.update_old_pi_op, self.update_v_target_op] if (
                self.t % g.update_target_v_interval == 0 and g.off_policy) else self.update_old_pi_op)

    def update_learning_rates_op(self):
        self.effective_timesteps_so_far += self.num_ts
        self.lr_mult = max(1.0 - float(self.effective_timesteps_so_far) / g.total_ts, 0.01)
        self.tactile_v_lr = g.v_lr * self.lr_mult
        self.tactile_q_lr = g.q_lr * self.lr_mult
        self.tactile_pi_lr = g.tactile_pi_lr * self.lr_mult

    def update_off_policy_buffer(self, buffer):
        if g.off_policy:
            buffer_size = len(list(buffer.queue))
            for i in range(buffer_size):
                batch = buffer.get()
                acts = self.sess.run(self.tactile_curr_pi['sampled_action'], feed_dict=batch)
                batch[self.tactile_act_off_ph] = acts
                buffer.put(batch)
        return buffer

    def prepare_learning(self, experience):
        self.compute_data(experience)
        self.pi_replay_buffer = Queue(maxsize=g.replay_buffer_queue_size)
        self.pi_replay_buffer = self.prepare_pi_replay_buffer(bs=g.tactile_pi_mini_bs, buffer=self.pi_replay_buffer)
        self.v_replay_buffer_pos, self.v_replay_buffer_neg, self.v_replay_buffer = self.prepare_v_replay_buffer(
            bs=g.tactile_v_mini_bs, pos_buffer=self.v_replay_buffer_pos, neg_buffer=self.v_replay_buffer_neg,
            buffer=self.v_replay_buffer)
        self.pi_replay_buffer = self.update_off_policy_buffer(self.pi_replay_buffer)
        self.v_replay_buffer = self.update_off_policy_buffer(self.v_replay_buffer)
        self.pi_replay_buffer_size = len(list(self.pi_replay_buffer.queue))
        self.v_replay_buffer_size = len(list(self.v_replay_buffer.queue))

        process = psutil.Process(os.getpid())
        self.memory = process.memory_percent()

        self.fd = {self.tactile_act_off_ph: self.on_acts, self.tactile_act_on_ph: self.on_acts,
                   self.obv_text_ph: str(self.obvs[-1]), self.act_text_ph: str(self.on_acts[0]),
                   self.tactile_pi_lr_ph: self.tactile_pi_lr, self.tactile_r_ph: self.rews,
                   self.tactile_v_lr_ph: self.tactile_v_lr, self.tactile_adv_ph: self.advs,
                   self.tactile_v_target_ph: self.vtarg, self.tactile_q_lr_ph: self.tactile_q_lr,
                   self.tactile_done_ph: self.dones, self.tactile_lr_mult_ph: self.lr_mult,
                   self.seen_objs_ph: g.seen_objs, self.novel_objs_ph: g.novel_objs,
                   self.success_rate_per_obj_ph: self.success_rate_per_obj,
                   self.success_rate_per_obj_values_ph: self.success_rate_per_obj_values,
                   self.restore_path_ph: str(g.path), self.raw_act_text_ph: str(self.infos[0]['raw_act']),
                   self.force_lift_ph: self.force_lifts, self.force_open_ph: self.force_opens}

        for name in g.tactile_obv_names:
            if getattr(g, 'tactile_%s' % name):
                self.fd[getattr(self, '%s_obv_ph' % name)] = getattr(self, '%s_obvs' % name)
                self.fd[getattr(self, 'delta_%s_obv_ph' % name)] = getattr(self, 'delta_%s_obvs' % name)
                if g.off_policy:
                    self.fd[getattr(self, '%s_next_obv_ph' % name)] = getattr(self, '%s_next_obvs' % name)

    def prepare_pi_replay_buffer(self, bs, buffer):
        for i in range(int(self.num_ts // bs)):
            real_ord = range(i * bs, min((i + 1) * bs, self.num_ts))

            batch_fd = {self.tactile_act_on_ph: self.on_acts[real_ord], self.tactile_done_ph: self.dones[real_ord],
                        self.tactile_adv_ph: self.advs[real_ord], self.tactile_v_target_ph: self.vtarg[real_ord],
                        self.tactile_r_ph: self.rews[real_ord], self.tactile_pi_lr_ph: self.tactile_pi_lr,
                        self.tactile_lr_mult_ph: self.lr_mult, self.tactile_q_lr_ph: self.tactile_q_lr,
                        self.tactile_v_lr_ph: self.tactile_v_lr, self.force_lift_ph: self.force_lifts[real_ord],
                        self.force_open_ph: self.force_opens[real_ord]}

            for name in g.tactile_obv_names:
                if getattr(g, 'tactile_%s' % name):
                    batch_fd[getattr(self, '%s_obv_ph' % name)] = getattr(self, '%s_obvs' % name)[real_ord]
                    batch_fd[getattr(self, 'delta_%s_obv_ph' % name)] = getattr(self, 'delta_%s_obvs' % name)[real_ord]
                    if g.off_policy:
                        batch_fd[getattr(self, '%s_next_obv_ph' % name)] = getattr(self, '%s_next_obvs' % name)[
                            real_ord]

            if buffer.full():
                buffer.get()
            buffer.put(batch_fd)
        return buffer

    def prepare_v_sub_buffer(self, buffer, indices):
        for i in indices:
            sample_fd = {self.tactile_act_on_ph: self.on_acts[i], self.tactile_done_ph: self.dones[i],
                         self.tactile_adv_ph: self.advs[i], self.tactile_v_target_ph: self.vtarg[i],
                         self.tactile_r_ph: self.rews[i], self.force_lift_ph: self.force_lifts[i],
                         self.force_open_ph: self.force_opens[i]}
            for name in g.tactile_obv_names:
                if getattr(g, 'tactile_%s' % name):
                    sample_fd[getattr(self, '%s_obv_ph' % name)] = getattr(self, '%s_obvs' % name)[i]
                    sample_fd[getattr(self, 'delta_%s_obv_ph' % name)] = getattr(self, 'delta_%s_obvs' % name)[i]
                    if g.off_policy:
                        sample_fd[getattr(self, '%s_next_obv_ph' % name)] = getattr(self, '%s_next_obvs' % name)[i]
            if buffer.full():
                buffer.get()
            buffer.put(sample_fd)
        return buffer

    def prepare_v_replay_buffer(self, bs, pos_buffer, neg_buffer, buffer):
        positive = self.vtarg > 0.5
        negative = 1 - positive
        pos_examples = [i for i in range(self.num_ts) if positive[i, 0]]
        neg_examples = [i for i in range(self.num_ts) if negative[i, 0]]
        pos_buffer = self.prepare_v_sub_buffer(pos_buffer, pos_examples)
        neg_buffer = self.prepare_v_sub_buffer(neg_buffer, neg_examples)

        while True:
            self.v_replay_buffer_pos_size = len(list(pos_buffer.queue))
            self.v_replay_buffer_neg_size = len(list(neg_buffer.queue))
            self.valid_buffer_size = min(self.v_replay_buffer_pos_size, self.v_replay_buffer_neg_size)
            if self.valid_buffer_size >= bs / 2:
                fd = {self.tactile_pi_lr_ph: self.tactile_pi_lr, self.tactile_lr_mult_ph: self.lr_mult,
                      self.tactile_q_lr_ph: self.tactile_q_lr, self.tactile_v_lr_ph: self.tactile_v_lr}
                pos_examples = []
                neg_examples = []
                for i in range(bs / 2):
                    pos_example = pos_buffer.get()
                    neg_example = neg_buffer.get()
                    pos_examples.append(pos_example)
                    neg_examples.append(neg_example)
                all_examples = pos_examples + neg_examples
                for key in dict(pos_example).keys():
                    fd[key] = [ex[key] for ex in all_examples]
                if buffer.full():
                    buffer.get()
                buffer.put(fd)
            else:
                break

        return pos_buffer, neg_buffer, buffer

    def compute_data(self, experience):
        obvs, on_acts, rews, next_obvs, dones, self.infos = zip(*experience)
        self.num_iter = int(self.num_ts // g.tactile_pi_mini_bs)

        # shuffle
        self.order = np.random.permutation(self.num_ts).tolist()

        on_acts = np.array(on_acts)
        rews = np.array(rews)
        dones = list(dones)
        fd = {}
        for name in g.tactile_obv_names:
            if getattr(g, 'tactile_%s' % name):
                fd[getattr(self, '%s_obv_ph' % name)] = np.array([obv[name] for obv in obvs])
                fd[getattr(self, 'delta_%s_obv_ph' % name)] = np.array([obv['delta_' + name] for obv in obvs])

        baseline = self.sess.run(self.tactile_curr_v_tr, feed_dict=fd)

        assert baseline.shape[0] == len(dones) == rews.shape[0], (len(dones), rews.shape, baseline.shape)
        seg = generate_seg(dones, rews, np.squeeze(baseline))
        add_vtarg_and_adv(seg)
        advs, vtarg = seg["adv"], seg["cum_r"]
        self.improved_tactile_variance = explained_variance(np.squeeze(baseline), np.squeeze(vtarg))

        self.advs = advs[self.order]
        self.vtarg = vtarg[self.order, np.newaxis]

        self.on_acts = on_acts[self.order]
        self.rews = rews[self.order]
        self.obvs = [obvs[i] for i in self.order]
        self.next_obvs = [next_obvs[i] for i in self.order]
        self.dones = np.array(dones)[self.order]
        self.infos = [self.infos[i] for i in self.order]
        self.force_lifts = np.array([info['forcelift'] for info in self.infos])
        self.force_opens = np.array([info['forceopen'] for info in self.infos])

        for name in g.tactile_obv_names:
            if getattr(g, 'tactile_%s' % name):
                setattr(self, '%s_obvs' % name, np.array([obv[name] for obv in self.obvs]))
                setattr(self, 'delta_%s_obvs' % name, np.array([obv['delta_' + name] for obv in self.obvs]))
                if g.off_policy:
                    setattr(self, '%s_next_obvs' % name, np.array([next_obv[name] for next_obv in self.next_obvs]))

        self.train_success_perc = np.mean([info['success'] for info in self.infos if 'success' in info])
        self.train_regrasp_success_perc = np.mean(
            [info['regrasp_success'] for info in self.infos if 'regrasp_success' in info])
        if self.t % g.check_tactile_baseline == 0:
            self.baseline_r = np.mean([obv['baseline'] for obv in self.obvs if 'baseline' in obv])
        self.outperformance = self.train_success_perc - self.baseline_r
        self.max_outperformance = max(self.max_outperformance, self.outperformance)
        self.max_train_success_perc = max(self.max_train_success_perc, self.train_success_perc)
        self.dynamic_tactile_delta = dynamicTactileDelta(self.max_train_success_perc)
        self.regrasp_penalty = np.mean([info['regrasp_penalty'] for info in self.infos])
        self.calculate_success_rate_per_obj()

    def demo(self):
        print('testing novel graspable_bodies...')
        if self.real:
            ray.init(num_cpus=g.num_actors, logging_level=logging.ERROR)
            module = __import__("algorithms.%s" % d.config.algo, fromlist=[d.config.algo])
            algo_class = ray.remote(getattr(module, d.config.algo))
            print('num_actors', g.num_actors)
            self.actors = [algo_class.remote(id + 1, True, self.dir_name, real=True) for id in range(g.num_actors)]
            ray.get([actor.restore_model_eval_op.remote() for actor in self.actors])
        while True:
            [os.remove(os.path.join(d.server_graspit_img_dir, i)) for i in os.listdir(d.server_graspit_img_dir)]
            if self.real:
                s = self.generate_grasp_multithreaded(test=True, multi=(not g.eval_single_object), novel=False,
                                                      num_objects=None)
            else:
                s = self.generate_grasp(test=True, multi=(not g.eval_single_object), novel=False, num_objects=None)
            done = False
            print('press to continue... start')
            state_id = s['state_id']
            try:
                t = 1
                while not done:
                    if s is not None:
                        print(s['positions'][-8 * g.check_history:])
                        print('palm', s['current_contacts'][:24].sum())
                        print('finger1', s['current_contacts'][24:48].sum())
                        print('finger2', s['current_contacts'][48:72].sum())
                        print('finger3', s['current_contacts'][72:].sum())

                    action, raw_act = self.get_tactile_action([s], best=True)
                    s, r, done, info = self.env.tactile_step(self.get_tactile_action_dict(action[0]),
                                                             info={'t': t, 'eval': True, 's': s, 'a': raw_act})
                    if not done and 'regrasp' in s:
                        s = self.generate_grasp(test=True, multi=(not g.eval_single_object), novel=False,
                                                num_objects=None, get_baseline=False, state_id=state_id)

                    t += 1
            except:
                traceback.print_exc()
                ipdb.set_trace()

    def build_tactile_histogram_ops(self):
        for name in g.tactile_obv_names:
            if getattr(g, 'tactile_%s' % name):
                tf.summary.histogram('%s_obv' % name, getattr(self, '%s_obv_ph' % name))
        tf.summary.histogram('tactile_act_bern', self.tactile_act_on_ph[:, :6])
        tf.summary.histogram('tactile_act_logit', self.tactile_act_on_ph[:, 6:])
        tf.summary.histogram('tactile_raw_act', self.tactile_curr_obv_tr)
        tf.summary.histogram('curr_v', self.tactile_curr_v_tr)
        tf.summary.histogram('tactile_r_ph', self.tactile_r_ph)
        tf.summary.histogram('tactile_adv_ph', self.tactile_adv_ph)
        if g.off_policy:
            tf.summary.histogram('tactile_v_target_ph', self.tactile_v_target_ph)
            tf.summary.histogram('q1', self.tactile_q1_on_tr)
            tf.summary.histogram('q2', self.tactile_q2_on_tr)
            tf.summary.histogram('next_v', self.tactile_next_v_target_tr)
            tf.summary.histogram('positive_mse_tr', self.positive_mse_tr)
            tf.summary.histogram('negative_mse_tr', self.negative_mse_tr)
        tf.summary.histogram('normalized_adv_tr', self.normalized_adv_tr)
        tf.summary.histogram('raw_adv_tr', self.raw_adv_tr)
        tf.summary.histogram('adv_mean_tr', self.adv_mean_tr)
        tf.summary.histogram('adv_var_tr', self.adv_var_tr)
        tf.summary.histogram('adv_std_tr', self.adv_std_tr)

    def build_tactile_text_ops(self):
        self.obv_text_ph = tf.placeholder(tf.string, name='obv_text_ph')
        self.act_text_ph = tf.placeholder(tf.string, name='act_text_ph')
        self.raw_act_text_ph = tf.placeholder(tf.string, name='raw_act_text_ph')
        tf.summary.text('obv', self.obv_text_ph)
        tf.summary.text('act', self.act_text_ph)
        tf.summary.text('raw_act', self.raw_act_text_ph)

    def generate_grasp_plan(self, s, post_point_cloud=False):
        if post_point_cloud:
            self.env.commander.post_point_cloud()
        t = 1
        done = False
        curr_t_steps = 0
        index = 0
        while True:
            try:
                while not done:
                    s['image_path'] = os.path.join(d.dirname, 'grasp_ts_%s_t_%s_id_%s.png' % (curr_t_steps, t, self.id))
                    cv2.imwrite(s['image_path'], s['image'])
                    action, obv = self.get_grasp_action(s)
                    a = self.get_grasp_action_dict(action[0])
                    if index is not None:
                        a['index'] = index
                    info = {'s': s, 't': t}
                    s, r, done, grasp_info = self.env.grasp_plan(a, info=info)
                    curr_t_steps += 1
                info['a'] = a
                return info
            except:
                if self.is_master():
                    traceback.print_exc()
                return None

    def generate_grasp_multithreaded(self, test, multi, novel, num_objects):
        self.env.init_world(test=test, multi=multi, novel=novel, num_objects=num_objects)
        s = self.env.reset()
        while self.need_reinit_pybullet(s):
            self.env.init_world(test=test, multi=multi, novel=novel, num_objects=num_objects)
            s = self.env.reset()
        plans = []
        trial = 0
        while not plans:
            slave_post_point_clouds = (trial % 10 == 0)
            if slave_post_point_clouds:
                self.env.commander.post_point_cloud()
            raw_plans = ray.get([actor.generate_grasp_plan.remote(s, slave_post_point_clouds) for actor in self.actors])
            plans = [plan for plan in raw_plans if plan is not None]
            trial += 1
        zs = [self.env.world_pos(plan['a'], s)[1][2] for plan in plans]
        print('z', zs)
        good_plan = random.choice(plans)
        while True:
            try:
                s, r, done, grasp_info = self.env.grasp_step(a=good_plan['a'], info=good_plan)
                break
            except:
                traceback.print_exc()
                pass
        assert done
        return s
