import copy

import ipdb

try:
    from queue import Queue
except:
    from Queue import Queue

from test_env.grasp_env import grasp_env
from utils import *


class tactile_env(grasp_env):
    def get_tactile_obv(self, start=False):
        if start:
            self.commander.contacts_queue = Queue(maxsize=g.history)
            self.commander.positions_queue = Queue(maxsize=g.history)
            self.commander.forces_queue = Queue(maxsize=g.history)
            self.commander.torques_queue = Queue(maxsize=g.history)
            self.commander.cartesians_queue = Queue(maxsize=g.history)
            self.commander.delta_contacts_queue = Queue(maxsize=g.history - 1)
            self.commander.delta_positions_queue = Queue(maxsize=g.history - 1)
            self.commander.delta_forces_queue = Queue(maxsize=g.history - 1)
            self.commander.delta_torques_queue = Queue(maxsize=g.history - 1)
            self.commander.delta_cartesians_queue = Queue(maxsize=g.history - 1)
            self.commander.full_cartesians_queue = Queue(maxsize=1000)
            for i in range(g.history):
                self.commander.contacts_queue.put(np.zeros(self.commander.num_tactile_links))
                self.commander.positions_queue.put(self.commander.getTactileJointPositions())
                self.commander.forces_queue.put(np.zeros_like(self.commander.getWristReactionForces()))
                self.commander.torques_queue.put(np.zeros_like(self.commander.getWristReactionTorques()))
                self.commander.cartesians_queue.put(np.zeros(self.commander.num_tactile_links * 3))
            for i in range(g.history - 1):
                self.commander.delta_contacts_queue.put(np.zeros(self.commander.num_tactile_links))
                self.commander.delta_positions_queue.put(np.zeros_like(self.commander.getTactileJointPositions()))
                self.commander.delta_forces_queue.put(np.zeros_like(self.commander.getWristReactionForces()))
                self.commander.delta_torques_queue.put(np.zeros_like(self.commander.getWristReactionTorques()))
                self.commander.delta_cartesians_queue.put(np.zeros(self.commander.num_tactile_links * 3))
        try:
            ret = self.commander.getForces()
            for name in g.tactile_obv_names:
                if getattr(g, 'tactile_%s' % name):
                    raw_data = np.array(list(getattr(self.commander, '%s_queue' % name).queue))
                    ret[name] = raw_data.flatten()
                    if g.tactile_delta:
                        delta = (raw_data[-1] - raw_data[-2]).flatten()
                        if name == 'positions':
                            delta = delta < g.delta_tolerance
                        if getattr(self.commander, 'delta_%s_queue' % name).full():
                            getattr(self.commander, 'delta_%s_queue' % name).get()
                        getattr(self.commander, 'delta_%s_queue' % name).put(delta.astype(np.float32))
                        ret['delta_' + name] = np.array(
                            getattr(self.commander, 'delta_%s_queue' % name).queue).flatten()

            return ret
        except:
            traceback.print_exc()
            ipdb.set_trace()

    def grasp_step(self, a, info=None, next_s=False, get_baseline=False):
        s = info['s']
        if (a['terminate'] and not g.zoom_image) or info['t'] >= g.grasp_horizon:
            default_env_state_id = self.commander.saveState()
            if get_baseline and not self.real and not g.record:
                state_id = self.grasp(s, a, grasp=True, get_pre_grasp_state=True)
                r = self.commander.computeQuality(delete=False)
                if self.eval:
                    raw_input(str({'baseline': r}))
                self.commander.restoreState(state_id)
            else:
                self.grasp(s, a, grasp=False)
            obv = self.get_tactile_obv(start=True)
            if get_baseline and not self.real and not g.record:
                obv['baseline'] = r
            else:
                obv['baseline'] = 0
            obv['state_id'] = default_env_state_id
            return obv, 0, True, {}
        else:
            try:
                image, transform = self.get_cropped_image(s['real_depth'], a, s['transform'], s['max_depth'])
            except:
                traceback.print_exc()
                copy_grasp_a = copy.copy(a)
                copy_grasp_a['terminate'] = True
                return self.grasp_step(copy_grasp_a, info)

            return {'image': image, 'transform': transform, 'real_depth': s['real_depth'], 'max_depth': s['max_depth'],
                    't': info['t'], 'rgb': s['rgb'] if next_s else None,
                    'scale': transform['dx'] * 2.0 / d.camera_w}, 0, False, {
                       'no_force': int(info['t'] < g.grasp_horizon)}

    def grasp_plan(self, a, info=None, next_s=False):
        s = info['s']
        if (a['terminate'] and not g.zoom_image) or info['t'] >= g.grasp_horizon:
            return self.grasp(s, a, grasp=False, execute=False), 0, True, {}
        else:
            try:
                image, transform = self.get_cropped_image(s['real_depth'], a, s['transform'], s['max_depth'])
            except:
                traceback.print_exc()
                copy_grasp_a = copy.copy(a)
                copy_grasp_a['terminate'] = True
                return self.grasp_plan(copy_grasp_a, info)

            return {'image': image, 'transform': transform, 'real_depth': s['real_depth'], 'max_depth': s['max_depth'],
                    't': info['t'], 'rgb': s['rgb'] if next_s else None,
                    'scale': transform['dx'] * 2.0 / d.camera_w}, 0, False, {
                       'no_force': int(info['t'] < g.grasp_horizon)}

    def deadlock(self, info):
        return np.allclose(info['s']['delta_positions'][-8 * g.check_history:], 1)

    def no_tactile_contacts(self, info):
        if max(info['s']['current_positions'][:3]) > 0.2:
            return 0
        else:
            return -1

    def planned_adjustment(self):
        nc = self.commander.num_cells_per_link
        current_hand_position = self.commander.getCurrentHandPose()
        try:
            cartesians = np.array(list(self.commander.full_cartesians_queue.queue)[::-1])
            if np.allclose(cartesians, 0):
                print('link_cartesians', [])
                return np.array([0, 0, 0])
            else:
                while True:
                    for cartesian in cartesians:
                        if not np.allclose(cartesian, 0):
                            reshaped = np.reshape(cartesian, (self.commander.num_tactile_links, 3))

                            link_cartesians = []
                            for i in range(len(self.commander.tactile_link_names)):
                                contacts = [c for c in reshaped[i * nc:(i + 1) * nc] if not np.allclose(c, 0)]
                                if len(contacts):
                                    mean_contact = np.mean(contacts, axis=0)
                                    link_cartesians.append(mean_contact)
                                    self.commander.postTactileCartesian(mean_contact + current_hand_position,
                                                                        name=self.commander.tactile_link_names[i])
                                else:
                                    self.commander.postTactileCartesian(current_hand_position,
                                                                        name=self.commander.tactile_link_names[i])

                            ret = np.mean(link_cartesians, axis=0)
                            print('link_cartesians', link_cartesians)
                            ret[2] = 0
                            normalized_ret = ret / np.linalg.norm(ret)
                            extra_ret = normalized_ret * g.additional_pos_adj
                            if self.eval:
                                print('extra_ret', extra_ret)
                            ret += extra_ret
                            self.commander.postTactileCartesian(ret + current_hand_position, name='adjustment')
                            return ret
        except:
            ipdb.set_trace()

    def tactile_step(self, a, info):
        force_open = self.deadlock(info)
        regrasp_action = g.regrasp and (a['regrasp'] or force_open)
        if g.teleoperation:
            a['regrasp'] = int(raw_input('regrasp?'))
            a['terminate'] = a['regrasp']
        if self.eval and not self.real and not g.record:
            raw_input((a, info['a']))
        elif self.real:
            print(a, info['a'])
        if info['t'] >= g.tactile_horizon:
            r = self.commander.computeQuality()
            return self.get_tactile_obv(), r, True, {'success': r, 'regrasp_penalty': 0, 'forcelift': True,
                                                     'forceopen': force_open}

        if (a['terminate'] and not a['regrasp'] and not g.demo_position) or ((a['terminate'] or regrasp_action) and g.demo_position):
            done = self.commander.incrementalLift()
            if done:
                r = self.commander.computeIncrementalQuality(delete=True)
                return self.get_tactile_obv(), r, True, {'success': r, 'regrasp_penalty': 0, 'forcelift': False,
                                                         'forceopen': force_open}
            else:
                return self.get_tactile_obv(), 0, False, {'regrasp_penalty': 0, 'forcelift': False,
                                                          'forceopen': force_open}
        elif regrasp_action:
            penalty_mult = self.no_tactile_contacts(info) if not force_open else 0
            if g.visual_regrasp:
                return {'visual_regrasp': True}, g.reopen_r_penalty * penalty_mult, False, {
                    'regrasp_penalty': g.reopen_r_penalty * penalty_mult, 'forcelift': False, 'forceopen': force_open}
            else:
                pos_angle = np.pi * float(a['pos_adjust_direction'])
                ori_angle = np.pi * float(a['ori_adjust_direction'])
                self.commander.dynamicOpen()
                if g.translation_planning:
                    pos_adjustment = self.planned_adjustment()
                else:
                    pos_adjustment = np.array([np.sin(pos_angle), np.cos(pos_angle), 0]) * g.adjust_distance * a[
                        'pos_adjust_distance']
                if self.eval and not g.record:
                    a = raw_input('before adjust %s %s' % (pos_adjustment, ori_angle))
                else:
                    a = 'y'
                self.commander.adjustBasePosition(pos_adjustment)
                self.commander.adjustHandOrientation(ori_angle, safe_lift=(a != 'n'))
                self.commander.avoidTrayCollision(lift=False)

                return self.get_tactile_obv(start=True), g.reopen_r_penalty * penalty_mult, False, {
                    'regrasp_penalty': g.reopen_r_penalty * penalty_mult, 'forcelift': False, 'tactile_regrasp': True,
                    'forceopen': force_open}
        else:
            self.commander.incrementalGrasp(finger1=a['finger1'], finger2=a['finger2'], finger3=a['finger3'],
                                            delta=a['delta'])

            return self.get_tactile_obv(), 0, False, {'regrasp_penalty': 0, 'forcelift': False, 'forceopen': force_open}
